///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  RefObj.cpp                                                                        //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#include    <iostream>
#include    <iomanip>
#include    <typeinfo>
#include    <map>
#include    "RefObj.h"
#include    "CharacterObj.h"
#include    "GrammarObj.h"
#include    "AndRuleObj.h"

using namespace std;


namespace  wej_parser
{
///////////////////////////////////////////////////////////////////////////////////////////////////
// string   RefObj::Display
///////////////////////////////////////////////////////////////////////////////////////////////////

string  RefObj::Display( bool , const unsigned int  ) const
{
  ostringstream        oS;
  oS << "REF(" << this->Link << ")";
  oS << DisplayFlags();
  return    oS.str();
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// bool   RefObj::Parse
///////////////////////////////////////////////////////////////////////////////////////////////////

bool  RefObj::Parse( ParseFrame PF )
{
  // Set rule
  PF.SetRule(this);

  if(GrammarAbs::DebugParse)
  {
//    cout << endl;
    cout << ">>REF"
         << "  \'" << this->Link << "\'"
         << "  [" << PF.GetRuleDebugDescriptor() << "]"
         << "  [" << PF.GetInPos() << "," << PF.GetOutListPos() << "]"
         << endl;
    cout.flush();
  }

  // Fixup Caches prior to execution at Parse or Inverse Parse
  if( !this->Cache )      this->FixupCachesPriorToExecutionAtParse();

  // Perform the target parse
  bool      R   = this->Cache->Parse(PF);

  if(R)     PF.AcceptAll();

  if(GrammarAbs::DebugParse)
  {
    cout  << "<<REF"
          << "  \'" << this->Link << "\'"
          << "  " << (R ? "<++++++>" : "<------>")
          << "  [" << PF.GetRuleDebugDescriptor() << "]"
          << "  [" << PF.GetInPos() << "," << PF.GetOutListPos() << "]"
          << endl << endl;
  }

  return    R;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// RefObj::GetOutput( ParseFrame& PF, const unsigned int RuleNum )
///////////////////////////////////////////////////////////////////////////////////////////////////

string  RefObj::GetOutput( ParseFrame& PF, const unsigned int RuleNum )
{
  return    PF.GetSegmentChars( RuleNum );
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// void   RefObj::GenerateCppParseTree
///////////////////////////////////////////////////////////////////////////////////////////////////

void    RefObj::GenerateCppParseTree( IBnfToCppCov& OutputData, bool KleinLayer, string& Includes )
{
  ostringstream     oS;
  UNUSED(Includes);

  oS << "  REF = new RefObj;" << endl;
  oS << "  REF->SetLink(\"" << this->Link << "\");" << endl;

  if( KleinLayer )
  {
    oS << "  KLEIN->SetKleinChild(REF);" << endl;
  }
  else
  {
    oS << "  AND->PushBack(REF);" << endl;
  }

  // Accumulate the output
  OutputData.Subroutines.back().Contents  += oS.str();
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// bool   RefObj:FixupCachesPriorToExecutionAtParse
///////////////////////////////////////////////////////////////////////////////////////////////////

bool    RefObj::FixupCachesPriorToExecutionAtParse(void)
{
  const GrammarAbs*                                       G;
  map<string,GrammarObj::MapRecType>::const_iterator      I;

  // Find the GrammarObj level
  for( G=this->GetFather() ; G && (typeid(*G) != typeid(GrammarObj)) ; G=G->GetFather() );

  // Cast pointer to a GrammarObj item
  const GrammarObj     *GR = dynamic_cast<const GrammarObj*>(G);

  // Look up the Label
  I   = GR->GetRuleMap().find(Link);

  // Did we found it
  if( I == GR->GetRuleMap().end() )
  {
    cout << "*Error:  Cannot find statement '" << Link << "'." << endl;
    return    false;
  }
  else
  {
    Cache   = I->second.Address;
  }

  // make sure the cache got defined
  assert(Cache);

  return  true;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// bool   RefObj::InverseParse
///////////////////////////////////////////////////////////////////////////////////////////////////

bool  RefObj::InverseParse( InverseParseFrame IPF )
{
  // Set rule
  IPF.SetRule(this);

  if( IPF.IsDebug() )
  {
    cout << ">> IParse REF[" << this->Link << ":  " << IPF.Display() << "]" << endl;
  }

  // Fix up Caches prior to execution at Parse or Inverse Parse
  if( !this->Cache )    this->FixupCachesPriorToExecutionAtParse();

  // Parse using the first rule
  auto  R   = this->Cache->InverseParse(IPF);

  // If we passed
  if(R)
  { // Accept the data on the streams
    IPF.AcceptAll();
  }

  if( IPF.IsDebug() )
  {
    const char    Ret = R ? '+' : '-';

    cout << "<<" << Ret << "IParse REF[" << this->Link << ":  " << IPF.Display() << "]" << endl;
  }

  return    R;
}
}   // End namespace 'wej_parser'


