///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  OverlapAndAddFiltering_001_lab.h                                                  //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//
//<LIBTAG:9749d6b8-f609-4b79-bdff-84550a04d39d>

#pragma once

#include  <fftw3.h>
#include  "AlignedMemAlloc.h"
#include  "ComplexHermitionVector.h"
#include  "RTIF.h"
#include  "FifoReaderTool.h"
#include  "FifoWriterTool.h"
#include  "OverlapAndAddFiltering_001_lab.h"


// For debug prints
extern  mutex  DebugMutex;     // Global Debug Mutex


class OverlapAndAddFiltering_001_lab : public RTIF
{
public:

  const bool    DebugPrints = false;

  typedef enum eProcessingThreadStartUpSeq
  {
    e_InitState,
    e_GotFilterCoefs,
    e_ReadFilterCoefs
  }   ProcessingThreadStartUpSeqType;

public:

  // Parameters
  const uint32_t    NumProcThreads_FromParams = 2;
  const uint32_t    NumChannels = 2;

  // Modified Parameters
  atomic<uint32_t>  NumProcThreads;
  uint32_t          MaxChannelsPerProcThread;

  // Threading and synchronization
  uint64_t          ThreadCycleCounter[2];   // Rolling cycle counter per thread
  uint32_t          ChannelStart[2];         // The starting channel number for each thread
  uint32_t          NumChannelsPerThread[2]; // The number of channels a particular thread processes
  thread            Thread[2];               // Thread handle for each thread
  thread            ControlThread;                                       // Thread handle for controller thread
  ProcessingThreadStartUpSeqType  StartUpState;
  mutex             FFTW_Mutex;                                          // Mutex used to serialize FFTW operations

  //////////////////////////////////////////////
  // Globally Available Data and Control Items
  //////////////////////////////////////////////

  // Input data   (_In)
  FIFO_Reader<int16_t,int16_t>  InData;
  FifoReaderTool<int16_t,int16_t>  FRT_In;

  // Filter Coefficients  (_Coefs)
  FIFO_Reader<complex<double>,complex<double>>  FirFilter;
  FifoReaderTool<complex<double>,complex<double>>  FRT_Coefs;

  // Output Filtered Data  (_Out)
  const vector<uint32_t>  PostFilt_Dimensions;
  const vector<bool>      PostFilt_WrapFlags;

  BufMap<int16_t,int16_t>   PostFilt_Extent;
  FIFO<int16_t,int16_t>     OutFifo;
  FifoWriterTool<int16_t,int16_t>  FWT_Out;

  ///////////////////////
  // Operational values
  //
  //   Advance:  The number of samples read from input
  //   FiltLen:  The number of coefficients in the filter
  //   FilterStartUpLag:  This number of samples are not output initially
  ////

  vector<uint32_t>  FilterStartUpLag;     // Is decremented during processing of batches
  uint32_t          CurrWriteAmount;      // Same for all threads
  vector<uint32_t>  READ;                 // Samples read per batch
  vector<uint32_t>  FiltLen;              // Number of coefficients
  vector<uint32_t>  FftLen;               // Length of FFT

  ////////////
  // Move to 'RenderedTemplateIF'
  ////////////

  bool        AnyActiveThreads(void) const;

  ////////////
  // Display
  ////////////

  string      Display_Constructor(void);
  string      Display_InitReset(void);
  string      Display_InitAlloc(void);
  string      Display_Kill(void);
  string      Display_Exit(void);

public:

  OverlapAndAddFiltering_001_lab(void);

  RTIF::ExitStateType  Init_Allocate(void);    // Called once to allocate dynamic memory
  RTIF::ExitStateType  Init_Reset(void);       // Called any time to reset all counters to their initial state
  RTIF::ExitStateType  KillThreads(void);      // Called to kill all threads
  RTIF::ExitStateType  Exit(void);             // Called to exit all threads

private:

  // Controller for the main processing threads
  static RTIF::ExitStateType   ProcessBatches( OverlapAndAddFiltering_001_lab& _Obj );   // Called to process a batch of data

  // Main processing thread entry point
  static RTIF::ExitStateType   ProcessingThread( OverlapAndAddFiltering_001_lab& _Obj, const uint32_t _ThreadNum );
};


