///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  LS_FIR_Filter.h                                                                   //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#pragma once

#include  <complex>
#include  <vector>
#include  <sstream>
#include  <list>
#include  <iomanip>
#include  <cassert>
#include  <gsl/gsl_sf.h>
#include  <fftw3.h>
#include  "AlignedMemAlloc.h"

using namespace std;


/**************************************************************************************************

  class  LS_FIR_Filter

  The ‘LS_FIR_Filter’ C++ class is a tool that generates the coefficients for a linear-phase FIR
  filter filters. These filters are specified in line segments of frequency and amplitude
  (i.e. freq:[F0,F1], amp: [A0,A1]).  Two types of segments: linear and exponential.  Linear
  segment’s amplitude are specified by an equation of the form B*f+A where B and A are picked
  so that the line-segment end points, freq:[F0,F1] and amp: [A0,A1], are satisfied.  The other
  type of segment is the exponential segment defined by exp(B*f+A).  In this case the end points
  are specified by the user as freq:[F0,F1], amp: [loge(A0),loge(A1)].  Please read the PDF that
  accompanies this software for examples of use and mathematical background.  These algorithms
  were developed and published by Amin G. Jaffer and William E. Jones in the 1994 timeframe.

  The Diagnostics member function gives an example of each segment type and the code associated
  with FIR filter coefficient generation and evaluation.

**************************************************************************************************/

class  LS_FIR_Filter
{
public:

  // Some constants
  static const double      TPI;
  static const double      PI2;

  // Segment types
  //  The interval is defined by [F0,F1] in frequency and [A0,A1] in amplitude for Linear Segments
  //  The interval is defined by [F0,F1] in frequency and [loge(A0),loge(A1)] in amplitude for Exponential Segments
  //  Note that frequencies are specified in the interval [0,1).  1/2 is PRF/2 at all sample rates.
  typedef enum  eSegType
  {
    eExp_RelErr,            // exp(B*f+A) segment, relative error
    eExp_AbsErr,            // exp(B*f+A) segment, absolute error
    eLin_RelErr,            // B*f+A segment, relative error
    eLin_AbsErr             // B*f+A segment, absolute error
  }   SegType;

private:

  // Exponential to a imaginary power
  static inline complex<double>   ExpI( const double _ImagPart )
  {
    return  complex<double>( cos(_ImagPart), sin(_ImagPart) );
  }

  // The E1 exponential integral
  static inline complex<double>   ExpInt_1_I( const double _ImagArg )
  {
    return  complex<double>( -gsl_sf_Ci(fabs(_ImagArg)), -.5*M_PI+gsl_sf_Si(_ImagArg) );
  }

  // The E2 exponential integral
  static inline complex<double>   ExpInt_2_I( const double _Arg )
  {
    return  ( ExpI(-_Arg) - MultI( _Arg * ExpInt_1_I( _Arg ) ) );
  }

  // Multiply by i
  static inline complex<double>   MultI( const complex<double> _Arg )
  {
    return  complex<double>( -imag(_Arg), real(_Arg) );
  }

public:

  /**************************************************************************************************
    class  Seg
  **************************************************************************************************/

  // A filter segment definition
  class   Seg
  {
  public:

    // Linear Segments are: B*f+A.  The interval is defined by [F0,F1] in frequency and [A0,A1] in amplitude.
    // Exponential Segments are: exp(B*f+A).  The interval is defined by [F0,F1] in frequency and
    //                                        [loge(A0),loge(A1)] in amplitude.
    //  Note that frequencies are specified in the interval [0,1).  1/2 is PRF/2 at all sample rates.
    SegType     Type;
    double      F0, F1, A0, A1, W;

    Seg( const SegType _Type, const double _F0, const double _F1, const double _A0, const double _A1, const double _W=1. )
    {
      Type      = _Type;
      F0        = _F0;
      F1        = _F1;
      A0        = _A0;
      A1        = _A1;
      W         = _W;
    }

    // This call modifies the Q matrix and R vectors to account for this segment.  It in turn calls
    //    specific generator functions based on the type of segment desired.
    //   Linear Segments are: B*f+A.  The interval is defined by [F0,F1] in frequency and [A0,A1] in amplitude.
    //   Exponential Segments are: exp(B*f+A).  The interval is defined by [F0,F1] in frequency and
    //                                          [loge(A0),loge(A1)] in amplitude.
    //  Note that frequencies are specified in the interval [0,1).  1/2 is PRF/2 at all sample rates.
    void    Gen_Segment( AVX_Vector<complex<double>>& Q,
                         AVX_Vector<complex<double>>& R, const double _Weight=1. );

    // Display a segment definition
    string   Display( const double SampleRate=1 ) const
    {
      ostringstream     oS;

      oS << "SEG[";

      switch(Type)
      {
      case eExp_RelErr:
        oS << "Exp_Rel";
      break;

      case eExp_AbsErr:
        oS << "Exp_Abs";
      break;

      case eLin_RelErr:
        oS << "Lin_Rel";
      break;

      case eLin_AbsErr:
        oS << "Abs_Lin";
      break;

      default:
        oS << "Error";
        assert(false);
      break;
      }

      oS  << "] "
          << setiosflags(ios::fixed) << setprecision(6) << setw(12) << (F0*SampleRate)
          << ": "
          << setiosflags(ios::fixed) << setprecision(8) << setw(12) << A0
          << "  <--->  "
          << setiosflags(ios::fixed) << setprecision(6) << setw(12) << (F1*SampleRate)
          << ": "
          << setiosflags(ios::fixed) << setprecision(8) << setw(12) << A1
          << "  W:"
          << setiosflags(ios::fixed) << setprecision(8) << setw(12) << W
          << "]";

      return    oS.str();
    }

  private:

    // exp(B*f+A) segment, relative error
    //  Note that frequencies are specified in the interval [0,1).  1/2 is PRF/2 at all sample rates.
    void    Gen_ExpRel_Segment( AVX_Vector<std::complex<double>>& Q,
                                AVX_Vector<std::complex<double>>& R,
                                const double _Weight  );

    // exp(B*f+A) segment, relative error
    //  Note that frequencies are specified in the interval [0,1).  1/2 is PRF/2 at all sample rates.
    void    Gen_ExpAbs_Segment( AVX_Vector<std::complex<double>>& Q,
                                AVX_Vector<std::complex<double>>& R,
                                const double _Weight  );

    // B*f+A segment, relative error
    //  Note that frequencies are specified in the interval [0,1).  1/2 is PRF/2 at all sample rates.
    void    Gen_LinRel_Segment( AVX_Vector<std::complex<double>>& Q,
                                AVX_Vector<std::complex<double>>& R,
                                const double _Weight  );

    // B*f+A segment, absolute error
    //  Note that frequencies are specified in the interval [0,1).  1/2 is PRF/2 at all sample rates.
    void    Gen_LinAbs_Segment( AVX_Vector<std::complex<double>>& Q,
                                AVX_Vector<std::complex<double>>& R,
                                const double _Weight  );
  };

private:

  // The segments list
  list<Seg>  Segs;

protected:

  // Linear system solution of 'matrix(Q) vector(Coefs) = vector(R)' via Levenson's
  bool    Levenson( AVX_Vector<complex<double>>& pR,
                    AVX_Vector<complex<double>>& g,
                    AVX_Vector<complex<double>>& X  );

public:

  // Add a new filter segment
  //   Linear Segments are: B*f+A.  The interval is defined by [F0,F1] in frequency and [A0,A1] in amplitude.
  //   Exponential Segments are: exp(B*f+A).  The interval is defined by [F0,F1] in frequency and
  //                                          [loge(A0),loge(A1)] in amplitude.
  //  Note that frequencies are specified in the interval [0,1).  1/2 is PRF/2 at all sample rates.
  void      Add( const SegType _Type, const double _F0, const double _F1, const double _A0, const double _A1, const double _Weight = 1. )
  {
    assert(_F0<_F1);
    Segs.push_back( Seg( _Type, _F0, _F1, _A0, _A1, _Weight ) );
  };

  // Add a new filter segment symmetric between the positive andnegative frequencies
  //   Linear Segments are: B*f+A.  The interval is defined by [F0,F1] in frequency and [A0,A1] in amplitude.
  //   Exponential Segments are: exp(B*f+A).  The interval is defined by [F0,F1] in frequency and
  //                                          [loge(A0),loge(A1)] in amplitude.
  //  Note that frequencies are specified in the interval [0,1).  1/2 is PRF/2 at all sample rates.
  void      AddSymmetric( const SegType _Type, const double _F0, const double _F1, const double _A0, const double _A1, const double _Weight = 1. )
  {
    assert(_F0<_F1);
    Segs.push_back( Seg( _Type,     _F0,     _F1, _A0, _A1, _Weight ) );
    Segs.push_back( Seg( _Type, 1.0-_F1, 1.0-_F0, _A1, _A0, _Weight ) );
  };

  // Generate the filter coefficients out into Out.  This involves the solution of the linear system
  //    'matrix(Q) vector(Coefs) = vector(R)'.
  bool  GenFilter( const unsigned int FilterLen, AVX_Vector<complex<double>>& Out );

  // Get the number of filter segments
  unsigned int  GetNumSegments(void) const
  {
    return  Segs.size();
  };

  // Clear all filter segments
  void  Clear(void)
  {
    Segs.clear();
  };

  // Get segment list reference
  list<Seg>&  GetSegs(void)
  {
    return    this->Segs;
  };
};


