///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  CharacterObj.h                                                                    //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#pragma once

#include    <cassert>
#include    <cstring>
#include    <list>
#include    "GrammarAbs.h"

using namespace std;


namespace   wej_parser
{
///////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////
// class   CharacterObj : public GrammarAbs
///////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////

class   CharacterObj : public GrammarAbs
{
public:

  ///////////////////////////////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////////////////////////
  // class   CharInt
  ///////////////////////////////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////////////////////////

  class   CharInt
  {
  public:

    unsigned int    Low, High;

  public:

    inline CharInt( const unsigned int pLow, const unsigned int pHigh )
    {
      Low     = pLow;
      High    = pHigh;
    }

    inline CharInt( const unsigned int pVal )
    {
      Low     = pVal;
      High    = pVal;
    }

    inline bool    operator==( const CharInt& Other ) const
    {
      return    Low == Other.Low;
    }

    inline bool    operator<( const CharInt& Other ) const
    {
      return    Low < Other.Low;
    }

    inline bool    Extend( const unsigned int Next )
    {
      if( Next != (High+1) )    return  false;
      High  = Next;
      return    true;
    }

    inline unsigned int   Size(void) const
    {
      return    High + 1 - Low;
    }

    inline string     DisplayAsInt(void) const
    {
      ostringstream   oS;
      oS << DisplayCharBnf(Low) << " to " << DisplayCharBnf(High);
      return    oS.str();
    }
  };

  ///////////////////////////////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////////////////////////
  // class   CharIntList
  ///////////////////////////////////////////////////////////////////////////////////////////////////
  ///////////////////////////////////////////////////////////////////////////////////////////////////

  class   CharIntList
  {
  public:

    list<CharInt>       Letters;

  public:

    inline void      Push( const unsigned int Val )
    {
      if( Letters.empty() )
      {
        Letters.push_back(CharInt(Val));
      }
      else if( Letters.back().Extend(Val) )
      {
        // NOOP
      }
      else
      {
        Letters.push_back(CharInt(Val));
      }
    }

    string          DisplayAsInt(void) const;
    unsigned int    Size(void) const;
  };

public:

  typedef  unsigned int          BitVectorNativeType;

  // Layout
  static const unsigned int      NativeBits    = sizeof(GrammarAbs::CharType) * 8;
  static const unsigned int      StorageBits   = sizeof(BitVectorNativeType)  * 8;
  static const unsigned int      NumChars      = 1 << NativeBits;
  static const unsigned int      NumNative     = (1 << (NativeBits - 3)) / sizeof(BitVectorNativeType);

  // Special Characters
  static const GrammarAbs::IntCharType        EOF_TOKEN  = NumChars;

  // Character graphics or encoding
  static string                               DisplayCharBnf( const GrammarAbs::IntCharType Char );
  static string                               DisplayCharCpp( const GrammarAbs::IntCharType Char );
  static string                               DisplayCharCpp( const string Str );

protected:

  // Access offsets and bit masks for the bit array
  inline static GrammarAbs::IntCharType       GetNativeItemNum( const GrammarAbs::IntCharType Letter )
  {
    return      Letter / StorageBits;
  }

  inline static BitVectorNativeType           GetNativeItemBit( const GrammarAbs::IntCharType Letter )
  {
    return      1 << (Letter % StorageBits);
  }

  // Get ordered char list
  CharIntList                                 GetCharList(void) const;
  CharIntList                                 GetInvertedCharList(void) const;

private:

  bool                        MatchEOF;
  BitVectorNativeType         CharArray[NumNative];

public:

  inline  CharacterObj(void)
  {
    Clear();
  }

  // Global operations
  void      Clear(void);
  void      Invert(void);

  inline void   InvertEofFlag(void)
  {
    MatchEOF    = !MatchEOF;
  }

  // Access Character Bits
  void                        SetChar( const GrammarAbs::IntCharType Char, const bool Value=true );
  BitVectorNativeType         GetChar( const GrammarAbs::IntCharType Char ) const;

  void                        SetRangeChars
                              (
                                const GrammarAbs::IntCharType     Char0,
                                const GrammarAbs::IntCharType     Char1,
                                const bool                        Value = true
                              );

  static string               DisplayXML( const string& S );

  //////////////////////////////
  // Interfaces
  //////////////////////////////

  string        Display( bool SubRule=false, const unsigned int Inden=0 ) const;
  bool          Parse( ParseFrame PF );
  string        GetOutput( ParseFrame& PF, const unsigned int RuleNum );
  void          GenerateCppParseTree( IBnfToCppCov& OutputData, bool KleinLayer, string& Includes );

  inline bool   DoesRead(void) const
  {
    return    true;
  }

  bool        InverseParse( InverseParseFrame IPF )
  {
    // Set rule
    IPF.SetRule(this);

    UNUSED(IPF);

    return  true;
  }
};
}   // End namespace 'wej_parser'


