///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  GenerateFiles.cpp                                                                 //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#include  <fstream>
#include  <sstream>
#include  <iomanip>
#include  "GenerateFiles.h"

using namespace std;

extern fs::path  PES_Tools_Root;


TemplateRenderer  Renderer;
list<RefRec>      Modules;
mutex             DebugMutex;


string    RefRec::Display(void) const
{
  ostringstream   oS;

  oS << ">> Module:  " << this->ModuleName << endl;

  for( auto I: ReferencedModules )
  {
    oS << "            --ref-->  " << I << endl;
  }

  return    oS.str();
}


string   RefRec::DisplayList( list<RefRec>& List )
{
  ostringstream   oS;

  for( auto I: List )
  {
    oS << I.Display();
  }

  return    oS.str();
}


// Reorder list so reference modules come first
void    RefRec::ReorderList( list<RefRec>& ModuleList )
{
  const bool  Debug = false;
  bool        Done  = true;

  // Make sure all the names in the 'ReferencedModules' are in the 'ModuleList' list
  bool    FoundMatch = false;

  for( auto I: ModuleList )
  {
    if(Debug)
    {
      cout << "[1a] " << I.ModuleName << endl;
    }

    for( auto J: I.ReferencedModules )
    {
      FoundMatch = false;

      if(Debug)
      {
        cout << "  [2a] " << J << endl;
      }

      // Make sure the referenced module is in the 'ModuleList' list
      for( auto K: ModuleList )
      {
        if( J == K.ModuleName )
        {
          FoundMatch = true;
          break;
        }
      }

      if(!FoundMatch)
      {
        Exception   E(__FILE__,__LINE__);
        E.Msg() << "Cannot find Referenced Module label '" << J << "'.";
        throw  E;
      }
    }
  }
  do
  {
    // Assume we're done till proven otherwise
    Done = true;

    // Walk through list of modules
    for( auto I=ModuleList.begin() ; I != ModuleList.end() ; ++I )
    {
      if(Debug)
      {
        cout << "[1b] " << I->GetModuleName() << endl;
      }

      // If there are no referenced items we don't need to worry
      if( I->ReferencedModules.size() == 0 )  continue;

      // Walk through list of references
      for( auto J=I->ReferencedModules.begin() ; J != I->ReferencedModules.end() ; ++J )
      {
        // Walk through list of modules
        for( auto K=ModuleList.begin() ; K != ModuleList.end() ; ++K )
        {
          // See if we found the reference before ourselves.  If so good.
          if( *J == K->ModuleName )   break;

          // See if we found ourselves.  If so that's bad and we need to reorder.
          if( I->ModuleName == K->ModuleName )
          {
            // There's a problem.  The reference must come after us.
            //   We move to the end of the list and we try again for ordering.
            RefRec  ToReorder = *I;
            ModuleList.erase(I);
            ModuleList.push_back(ToReorder);
            Done = false;
            break;
          }

          // We could not find the reference module
          if( J == I->ReferencedModules.end() )
          {
            Exception   E(__FILE__,__LINE__);
            E.Msg() << "Cannot file module '" << *J << "' referenced in '" << I->ModuleName <<"'.";
            throw  E;
          }
        }

        // Break out again and check the new ordering
        if(!Done)  break;
      }

      // Break out again and check the new ordering
      if(!Done)  break;
    }
  } while(!Done);
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// void    Init_StartCPP_File( const string& _OutputPath, const chrono::sys_time& _NowTime )
///////////////////////////////////////////////////////////////////////////////////////////////////

void    Init_StartCPP_File( const string& _OutputPath, const string& _NowTime )
{
  const string  FullOutputPath = _OutputPath + "/" + "start.cpp";

  ofstream  File( FullOutputPath, ios::trunc | ios::binary );

  if( !File.is_open() )
  {
    throw  Exception(string("Cannot create output file '" + FullOutputPath), __FILE__, __LINE__ );
  }

  // Declare the modules
  File << CreateTimestamp(_NowTime) << endl;

  File << "// Module includes" << endl;
  File << "#include  <fftw3.h>" << endl;
  File << "#include  <mutex>" << endl;
  File << "#include  <malloc.h>" << endl;
  File << "#include  \"modules.h\"" << endl;
  File << "#include  \"Exception.h\"" << endl << endl;

  File << "mutex  FFTW_Mut;" << endl << endl;

  // The 'main' routine
  File << "int    main( int /* NumArgs */, char** /* ArgList */ )" << endl << "{" << endl;

  File << "  assert( mallopt( M_CHECK_ACTION, 0 ) == 1 );" << endl << endl;

  File << "  try" << endl;
  File << "  {" << endl;

  // Init FFTW3 multi-threaded version
  File << "    {" << endl;
  File << "      lock_guard<mutex>  Lock(FFTW_Mut);" << endl;
  File << "      auto Flag = fftw_init_threads();" << endl;
  File << "      if( !Flag )" << endl;
  File << "      {" << endl;
  File << "        Exception  E(__FILE__,__LINE__);" << endl;
  File << "        E.Display() = \"Error executing fftw_init_threads().\";" << endl;
  File << "      }" << endl;
  File << "    }" << endl << endl;

  // Call 'Init_Alloc'
  File << "  ///////////////////////////" << endl;
  File << "  // Init_Allocate" << endl;
  File << "  ///////////////////////////" << endl << endl;

  for( RefRec& I: Modules )
  {
    File << "  " << I.GetModuleName() << "_dec.Init_Allocate();" << endl;
  }

  // Call 'Init_Alloc'
  File << endl << "  ///////////////////////////" << endl;
  File         << "  // Init_Reset" << endl;
  File         << "  ///////////////////////////" << endl << endl;

  for( RefRec& I: Modules )
  {
    File << "  " << I.GetModuleName() << "_dec.Init_Reset();" << endl;
  }

  File << endl;
  File << "  cout << endl << \"=====>  STARTING REAL-TIME OPERATION  <=====\" << endl << endl;" << endl;
  File << "  cout.flush();" << endl;

  // Debug sleep
  File << endl << "  //////////////////////////////" << endl;
  File         << "  // Debug sleep in background" << endl;
  File         << "  //////////////////////////////" << endl << endl;

  File << "  std::this_thread::sleep_for(std::chrono::seconds(14s));" << endl << endl;

  File << "  DebugMutex.lock();" << endl;
  File << "  cout << endl << \"=====> Got Kill <=====\" << endl;" << endl;
  File << "  DebugMutex.unlock();" << endl << endl;

  // Kill threads
  File << "  ///////////////////////////" << endl;
  File << "  // KillThreads" << endl;
  File << "  ///////////////////////////" << endl << endl;

  // Kill all the spawned processes in reverse order
  list<RefRec>::iterator  R;

  for( R=Modules.begin() ; R != Modules.end() ; ++R )
  {
    File << "  " << R->GetModuleName() << "_dec.KillThreads();" << endl;
  }

  File << endl << "  DebugMutex.lock();" << endl;
  File << "  cout << endl << "
       << endl
         << "    \"===> All the threads are dead and joined back up --- Program Terminated Normally \" "
       << endl << "    \"<===\" << endl << endl;"
       << endl;
  File << "  DebugMutex.unlock();" << endl;
  File << "  }" << endl << endl;

  File << "  catch(Exception&  DE)" << endl;
  File << "  {" << endl;
  File << "    cout << endl << DE.Display() << endl << endl;" << endl;
  File << "    return  1;" << endl;
  File << "  }" << endl << endl;

  File << "  return  0;" << endl;
  File << "}" << endl << endl;

  File.flush();
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// void    Init_ModulesCPP_File( const string& _OutputPath, const chrono::sys_time& _NowTime )
///////////////////////////////////////////////////////////////////////////////////////////////////

void    Init_ModulesCPP_File( const string& _OutputPath, const string& _NowTime )
{
  const string  FullOutputPath = _OutputPath + "/" + "modules.cpp";

  ofstream  File( FullOutputPath, ios::trunc | ios::binary );

  if( !File.is_open() )
  {
    throw  Exception(string("Cannot create output file '" + FullOutputPath + "'"), __FILE__, __LINE__ );
  }

  // Declare the modules
  File << CreateTimestamp(_NowTime) << endl;

  // Declare the modules
  File << "// Modules" << endl;

  File << "#include  \"modules.h" << "\"" << endl;

  File << endl << endl;

  for( RefRec& I: Modules )
  {
    File << I.GetModuleName() << "    " << I.GetModuleName() << "_dec;" << endl;
  }

  File << endl << endl;

  File.flush();
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// void    Init_ModulesH_File( const string& _OutputPath, const chrono::sys_time& _NowTime )
///////////////////////////////////////////////////////////////////////////////////////////////////

void    Init_ModulesH_File( const string& _OutputPath, const string& _NowTime )
{
  const string  FullOutputPath = _OutputPath + "/" + "modules.h";

  ofstream  File( FullOutputPath, ios::trunc | ios::binary );

  if( !File.is_open() )
  {
    throw  Exception(string("Cannot create output file '" + FullOutputPath + "'"), __FILE__, __LINE__ );
  }

  // Declare the modules
  File << CreateTimestamp(_NowTime) << endl;

  // Declare the modules
  File << "// Modules" << endl;

  for( RefRec& I: Modules )
  {
    File << "#include  \"" << I.GetModuleName() << ".h" << "\"" << endl;
  }

  File << endl << endl;

  for( RefRec& I: Modules )
  {
    File << "extern  " << I.GetModuleName() << "    " << I.GetModuleName() << "_dec;" << endl;
  }

  File << endl << endl;

  File.flush();
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// ProcessTemplates
///////////////////////////////////////////////////////////////////////////////////////////////////

void  ProcessTemplates
      (
        DataNodeRef     ModuleDef,
        const string&   OutputPath,
        const string&   NowTime
      )
{
  const bool  Debug = false;
  auto        TimeStart =  chrono::high_resolution_clock::now();

  // Recover the path to the output directory
  DataRef<string>  RootDirRef( DataPath("//Library/root_dir") );
  DataRef<string>  GenDateRef( DataPath("//Library/gen_date") );

  fs::path  LibPath( fs::path(RootDirRef.GetDataAsString()) / fs::path("lib") );
  fs::path  OutPath( OutputPath );

  // Recover the node tag
  const DataPath  DP_NodeTag( ModuleDef.GetDataPath() );
  const string    DP_NodeTag_PathString( ModuleDef.GetDataPath().GetPathString() );
  const auto      DP_NodeTag_PathString_Last = DP_NodeTag_PathString.find_last_of("/");
  const string    DP_NodeTag_PathStringSuffix( DP_NodeTag_PathString.substr( DP_NodeTag_PathString_Last+1 ) );

  if(Debug)
  {
    cout << "--" << ModuleDef.GetDataPath() << endl;
//    cout << "--[LibPath]         " << LibPath << endl;
//    cout << "--[OutPath]         " << OutPath << endl;
//    cout << "--[RootDirRef]      " << RootDirRef.GetDataAsString() << endl;
//    cout << "--[GenDateRef]      " << GenDateRef.GetDataAsString() << endl;
//    cout << "--[LibPath]         " << LibPath << endl;
//    cout << "--[OutPath]         " << OutPath << endl;
//    cout << "--[DP_NodeTag_PathStringSuffix]  " << DP_NodeTag_PathStringSuffix << endl;
  }

  const string  LibraryLabel( DP_NodeTag_PathStringSuffix );

  DataRef<string>  LabelTemplateUuid( ModuleDef.GetDataPath() + DataPath("uuid") );

  // Data Path
  DataPath  TemplateFilesPath = DataPath("//Library/lib") +
                                DataPath( string("u_") + LabelTemplateUuid.GetData() ) +
                                DataPath( string("files") );

  // Examine all files that use this UUID
  DataNodeRef  LibEntries( TemplateFilesPath , true );

  for( LibEntries.Begin() ; LibEntries.MoreChildren() ; LibEntries.Next() )
  {
    ProcessTemplate
    (
      LibPath,
      OutPath,
      ModuleDef,
      LibEntries,
      LibEntries.Curr(),
      NowTime
    );
  }

  if(true)
  {
    auto  TimeEnd = chrono::high_resolution_clock::now();
    std::chrono::duration<double>  ElapsedSeconds = TimeEnd - TimeStart;
    cout << ">>[Processed Template " << setw(8) << ElapsedSeconds.count() << "s]   "
         << ModuleDef.GetDataPath() << endl;
    cout.flush();
  }
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// ProcessTemplate
///////////////////////////////////////////////////////////////////////////////////////////////////

void    ProcessTemplate
        (
          const fs::path&  LibPath,
          const fs::path&  OutPath,
          DataNodeRef      ProgramDef,          // PROGRAM.XML being processed
          DataNodeRef      LibManifestRecDef,   // LIB_MANIFEST.XML: all files for uuid
          DataNodeRef      LibManifestFileDef,  // LIB_MANIFEST.XML: a particular file
          const string&    NowTime
        )
{
  const bool  Debug = false;

  // Recover the path to the output directory
  DataRef<string>  RootDirRef( DataPath("//Library/root_dir") );
  DataRef<string>  GenDateRef( DataPath("//Library/gen_date") );

  // Get the stem file name
  const string     StemFileName( ProgramDef.Curr().GetXmlNode().GetData() );
  DataRef<string>  RootDir(RootDirRef);
  DataRef<bool>    InitFlagRef( LibManifestFileDef.GetDataPath() + DataPath("init") );
  DataRef<string>  FilePathString( LibManifestFileDef.GetDataPath() + DataPath("path") );
  const auto       DotStringPos = FilePathString.GetData().find_last_of('.');
  const string     FileSuffix = FilePathString.GetData().substr( DotStringPos+1 );
  const fs::path   TargetFilePath_1( StemFileName + '.' + FileSuffix );
  const fs::path   TargetFilePath( OutPath / TargetFilePath_1 );
  const fs::path   SourceFilePath_1( FilePathString.GetData() );
  const fs::path   SourceLibPath ( LibPath / SourceFilePath_1 );

  // Realtime display
  cout << "  Generating Output File \'" << StemFileName << "." << FileSuffix << "\'" << endl;  cout.flush();

  if(Debug)
  {
    cout << endl;
    cout << ">>[ProcessTemplate]     " << ProgramDef.GetDataPath() << endl;

    cout << "--[RootDir]             " << boolalpha << RootDir.GetData() << endl;  cout.flush();
    cout << "--[LibPath]             " << LibPath << endl;
    cout << "--[OutPath]             " << OutPath << endl;
    cout << "--[GenDateRef]          " << GenDateRef.GetDataAsString() << endl;  cout.flush();
    cout << "--[StemFileName/LABEL]  " << StemFileName << endl;  cout.flush();
    cout << "--[InitFlagRef]         " << boolalpha << InitFlagRef.GetData() << endl;  cout.flush();
    cout << "--[FilePathString]      " << FilePathString.GetData() << endl;  cout.flush();
    cout << "--[FileSuffix]          " << FileSuffix << endl;  cout.flush();
    cout << "--[ProgramDef]        " << ProgramDef.Display(2);
    cout << "--[LibManifestRecDef] " << LibManifestRecDef.Display(2);
    cout << "--[LibManifestFileDef]" << LibManifestFileDef.Display(2);
    cout << "--[NowTime]             " << NowTime << endl;
    cout << "--[SourceLibPath]       " << SourceLibPath << endl;
    cout << "--[TargetFilePath]      " << TargetFilePath << endl;
  }

  // Module init flag used at startup
  if( InitFlagRef.GetData() )
  {
    Modules.push_back( RefRec(StemFileName) );

    /////////////////////////////////
    // Get references for this node
    /////////////////////////////////

    DataPath     DP( ProgramDef.GetDataPath() + DataPath("/referenced_modules") );
    DataNodeRef  NR( DP.GetPathString(), true );

    if( !NR.DoesNodeExist() )
    {
      throw   Exception
              (
                string("Data path '") + DP.GetPathString() + "' does not exist.",
                  __FILE__, __LINE__ );
    }

    for( NR.Begin() ; NR.MoreChildren() ; NR.Next() )
    { // Add the new reference
      const string    LabelToMatch{ NR.Curr().GetXmlNode().GetData() };
      const DataPath  TagLoc = LookupTagFromLabel(LabelToMatch);

      Modules.back().ReferencesList().push_back(LabelToMatch);

      if(Debug)
      {
        cout << "--->[LOOK_UP] " << LabelToMatch << " ---> " << TagLoc << endl;
      }
    }
  }

  /////////////////////////////////////
  // Perform XML Decoration Operation
  /////////////////////////////////////

  ifstream  IN( SourceLibPath, ios::binary );

  if( !IN.is_open() )
  {
    ostringstream  oS;
    oS << "Cannot open file '" << SourceLibPath << "'.";
    throw  Exception( oS.str(), __FILE__, __LINE__ );
  }

  // Parse the marked-up source file
  Renderer.Parse(IN);

  string  Out;

  Out += Renderer.ProcessOutput( ProgramDef.GetDataPath() );

  // Create the output file and write output it
  ofstream  OutFile( TargetFilePath, ios::trunc | ios::binary );

  if( !OutFile.is_open() )
  {
    ostringstream  oS;
    oS << "Cannot create output file '" << TargetFilePath << "'";
    throw  Exception( oS.str(), __FILE__, __LINE__ );
  }

  // Declare the modules
  OutFile << CreateTimestamp
             (
               NowTime,
               SourceLibPath,
               ProgramDef.GetDataPath().GetPathString()
             ) << endl;

  // Write the output to file
  OutFile.write( Out.c_str(), Out.size() );
  OutFile.flush();

  if(Debug)
  {
    cout << "<<[ProcessTemplate]     " << ProgramDef.GetDataPath() << endl;
    cout.flush();
  }
}


DataPath  LookupTagFromLabel( const string& LabelParam )
{
  const bool  Debug = false;

  if(Debug)
  {
    cout << "---[LookupTagFromLabel] " << LabelParam << endl;
  }

  DataPath     LibrariesPath("//Program/Modules");
  DataNodeRef  LibRef( LibrariesPath, true );

  if( !LibRef.DoesNodeExist() )
  {
    throw   Exception
            (
              string("Data path '") + LibrariesPath.GetPathString() + "' does not exist.",
                __FILE__, __LINE__ );
  }

  for( LibRef.Begin() ; LibRef.MoreChildren() ; LibRef.Next() )
  { // Add the new reference
    const string     LabelPath( LibRef.Curr().GetDataPath().GetPathString() + "/label" );
    DataPath         LabelP(LabelPath);
    DataRef<string>  LabelVal(LabelP);

    if(Debug)
    {
      cout << "-----[CMP: " << LabelParam << "  <--?-->  " << LabelVal.GetData() << "]" << endl;
    }

    if( LabelParam == LabelVal.GetData() )
    { // Found it.  Path to matched object.
      if(Debug)
      {
        cout << "----[FOUND: " << LibRef.Curr().GetDataPath().GetPathString() << "]" << endl;
      }

      return  LibRef.Curr().GetDataPath().GetPathString();
    }
  }

  if(Debug)
  {
    cout << "---[ERROR: Cannot find '" << LabelParam << "']" << endl;
  }

  return  DataPath("//Error/CANNOT_FIND");
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// CreateTimestamp
///////////////////////////////////////////////////////////////////////////////////////////////////

string  CreateTimestamp( const string& NowTime, const string Source, const string Definition )
{
  stringstream  Out;

  Out << "/*##################################################################################################" << endl;
  Out << "## Data Definition: '" << Definition << "'" << endl;
  Out << "## Source: '" << Source << "'" << endl;
  Out << "## Rendered by 'xml_decorator' at " << NowTime << "." << endl;
  Out << "## Copyright 2023, William Earl Jones, All Rights Reserved    [info: http://www.wejc.com]" << endl;
  Out << "##################################################################################################*/" << endl;

  return  Out.str();
}


