///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  GrammarObj.cpp                                                                    //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#include    <cassert>
#include    <iostream>
#include    <iomanip>
#include    <sstream>
#include    <typeinfo>
#include    "GrammarObj.h"
#include    "CharacterObj.h"
#include    "BNFC_Gen.h"
#include    "KleinWrapperObj.h"
#include    "RefObj.h"
#include    "LiteralObj.h"
#include    "OutputListObj.h"
#include    "AndRuleObj.h"
#include    "OrRuleObj.h"
#include    "TagObj.h"
#include    "XeqObj.h"
#include      "Timer.h"

using namespace std;


namespace   wej_parser
{
const map<string,int>   GrammarObj::TokenToRuleMap
{
  { "PARSER",           GrammarObj::PARSER_Token },
  { "XEQ",              GrammarObj::XEQ_Token },
  { "PUMP",             GrammarObj::PUMP_Token },
  { "STMNT",            GrammarObj::STMNT_Token },
  { "AND",              GrammarObj::AND_Token },
  { "REF_CMND",         GrammarObj::REF_CMND_Token },
  { "KLEIN",            GrammarObj::KLEIN_Token },
  { "OUT_LIST_START",   GrammarObj::OUT_LIST_START_Token },
  { "OUT_LIST_A_ITEM",  GrammarObj::OUT_LIST_A_ITEM_Token },
  { "OUT_LIST_A_CHAR",  GrammarObj::OUT_LIST_A_CHAR_Token },
  { "OUT_LIST_EL",      GrammarObj::OUT_LIST_EL_Token },
  { "CHAR_START",       GrammarObj::CHAR_START_Token },
  { "CHAR",             GrammarObj::CHAR_Token },
  { "CHAR_SINGLE",      GrammarObj::CHAR_SINGLE_Token },
  { "CHAR_RANGE",       GrammarObj::CHAR_RANGE_Token },
  { "CHAR_ALL",         GrammarObj::CHAR_ALL_Token },
  { "CHAREOF",          GrammarObj::CHAREOF_Token },
  { "CHAR_NOT_END",     GrammarObj::CHAR_NOT_END_Token },
  { "HEXCHAR",          GrammarObj::HEXCHAR_Token },
  { "LITERAL",          GrammarObj::LITERAL_Token },
  { "LIT_START",        GrammarObj::LIT_START_Token },
  { "LIT_EL",           GrammarObj::LIT_EL_Token },
  { "LIT_HEX2",         GrammarObj::LIT_HEX2_Token },
  { "LIT_END",          GrammarObj::LIT_END_Token },
  { "TAG",              GrammarObj::TAG_Token },
  { "CHAR_NOT_END",     GrammarObj::CHAR_NOT_END_Token }
};


///////////////////////////////////////////////////////////////////////////////////////////////////
// GrammarObj::~GrammarObj()
///////////////////////////////////////////////////////////////////////////////////////////////////

GrammarObj::~GrammarObj()
{
  unsigned int    K;

  for( K=0 ; K < Group.size() ; ++K )
  {
    if(Group[K])    delete Group[K];
  }
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// string   GrammarObj::Display
///////////////////////////////////////////////////////////////////////////////////////////////////

string    GrammarObj::Display( bool , const unsigned int Indent ) const
{
  ostringstream   oS;
  unsigned int    K, L;

  for( K=0 ; K<Indent ; ++K )   oS << ' ';
  oS << this->GetLabel() << ":" << endl;

  for( K=0 ; K<Indent ; ++K )   oS << ' ';
  oS << "{" << endl;

  for( L=0 ; L < Group.size() ; ++L )
  {
    if( L > 0 )    oS << endl;
    oS << Group[L]->Display(false,Indent);
  }

  for( K=0 ; K<Indent ; ++K )   oS << ' ';
  oS << "}" << endl;

  oS << DisplayFlags();

  return    oS.str();
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// bool   GrammarObj::ParseRule
///////////////////////////////////////////////////////////////////////////////////////////////////

bool    GrammarObj::ParseRule( GrammarAbs* StartRule, ParseFrame& PF )
{
  bool                      R;
  static wej_util::Timer    ParseTimer;

  if(this->ParseTime)
  {
    ParseTimer.Start();
  }

  // Set rule
  PF.SetRule(this);

  // If Debug
  if( PF.IsDebug() )
  {
    cout << ">>Parse GRAM[" << this->GetLabel() << ": "  << PF.Display() << "]" << endl;
    cout.flush();
  }

  assert(StartRule);

  // Parse using the first rule
  R   = StartRule->Parse(PF);

  // If we passed
  if(R)
  { // Accept the data on the streams
    PF.AcceptAll();
  }

  if(this->ParseTime)
  {
    this->ParseTimeSecs   = ParseTimer.GetCurrDelTimeSecs();
    this->ParseChars      = PF.GetInPos();

    this->ParseCharsPerSecond   = this->ParseChars;
    this->ParseCharsPerSecond  /= this->ParseTimeSecs;

    cout << "[T] PARSE RATE:  ";
    cout << setw(7)
         << setiosflags(ios::fixed)
         << setprecision(2)
         << (this->ParseCharsPerSecond / 1024.);
    cout << " KBytes/s  ";
    cout << setw(8)
         << setiosflags(ios::fixed)
         << setprecision(3)
         << (1000. * this->ParseTimeSecs);
    cout << " ms  ";
    cout << setw(8)
         << setiosflags(ios::fixed)
         << setprecision(3)
         << (this->ParseChars / 1024.);
    cout << " KBytes" << endl;
  }

  if(GrammarAbs::DebugParse)
  {
    cout << ">>Parse GRAM[" << PF.GetInPos() << "," << PF.GetOutListPos() << "]"
         << GetLabel() << ":  "
         << "R="
         << R
         << endl;
  }

  return    R;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// bool   GrammarObj::InverseParse
///////////////////////////////////////////////////////////////////////////////////////////////////

bool    GrammarObj::ParseRule( const string& StartRuleLabel, ParseFrame& PF )
{
  // Look up the rule name
  map<const string,MapRecType>::iterator    I;

  // Look up the label
  I  = this->Rules.find(StartRuleLabel);

  // If we didn't find the label
  if( I == this->Rules.cend() )  return  false;

  // Parse the file
  return    ParseRule( I->second.Address, PF );
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// bool     GrammarObj::InverseParse
///////////////////////////////////////////////////////////////////////////////////////////////////

bool    GrammarObj::InverseParse( InverseParseFrame IPF )
{
  // Set rule
  IPF.SetRule(this);

  // If Debug
  if( IPF.IsDebug() )
  {
    cout << ">> IParse GRAM[" << this->GetLabel() << ":  " << IPF.Display() << "]" << endl;
  }

  assert(!Group.empty());

  // Parse using the first rule
  auto  R   = Group.front()->InverseParse(IPF);

  // If we passed
  if(R)
  { // Accept the data on the streams
    IPF.AcceptAll();
  }

  // If Debug
  if( IPF.IsDebug() )
  {
    const char    Ret = R ? '+' : '-';

    cout << "<<" << Ret << "IParse GRAM[" << this->GetLabel() << ":  " << IPF.Display() << "]" << endl;
  }

  return    R;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// string     GrammarObj::GetOutput
///////////////////////////////////////////////////////////////////////////////////////////////////

string     GrammarObj::GetOutput( ParseFrame& PF, const unsigned int RuleNum )
{
  UNUSED(PF);
  UNUSED(RuleNum);

  return    string();
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// GrammarObj*       GrammarObj::CreateGrammarFromBnfcSource
///////////////////////////////////////////////////////////////////////////////////////////////////

GrammarObj*       GrammarObj::CreateGrammarFromBnfcSource( istream& In )
{
  list<OutputRec>::const_iterator   I;
  Outputs                           OUT_List;
  GrammarAbs*                       GA  = wej_parser::wej_bnf::Parse_BNFC();
  GrammarObj*                       G   = nullptr;
  OrRuleObj*                        OR  = nullptr;
  AndRuleObj*                       AND = nullptr;
  KleinWrapperObj*                  K   = nullptr;
  RefObj*                           R   = nullptr;
  OutputListObj*                    O   = nullptr;
  CharacterObj*                     C   = nullptr;
  LiteralObj*                       L   = nullptr;
  TagObj*                           T   = nullptr;
  XeqObj*                           X   = nullptr;
  GrammarAbs::IntCharType           C1=0, C2=0;
  string                            Literal;

  // Grab the data from the BNFC file
  {
    ParseFrame    PF( In, OUT_List );

    bool      Ret = GA->Parse(PF);
    delete    GA;

    if(!Ret)    return  nullptr;

    PF.AcceptAll();
  }

  // Over the output records
  for( I = OUT_List.GetDataList().begin() ; I != OUT_List.GetDataList().end() ; ++I )
  {
    // Ignore any debug records as they are for doc and test only
    if( I->IsDebug() )
    {
      cout << ">>DEBUG: " << I->GetDataVect()[0] << endl;
      continue;
    }

    // DEBUG
    if(false)
    {
      cout << setiosflags(ios::left) << setw(16) << I->GetDataVect()[0];

      for( unsigned int L=1 ; L < I->GetDataVect().size() ; ++L )
      {
        cout << "  ";
        cout << setiosflags(ios::left) << setw(16) << I->GetDataVect()[L];
      }

      cout << endl;
    }

    auto  RuleNum = GrammarObj::TokenToRuleMap.find( I->GetDataVect()[0] );

    if( RuleNum == GrammarObj::TokenToRuleMap.end() )
    {
      cout << "*Error: Cannot find GrammarObj Rule '" << I->GetDataVect()[0] << "'" << endl;
      return    nullptr;
    }

    switch( RuleNum->second )
    {
      case  PARSER_Token:
      {
        assert( I->GetDataVect().size() == 2 );
        assert( !G );

        G   = new GrammarObj;
        G->SetLabel(I->GetDataVect()[1]);
        break;
      }

      case  XEQ_Token:
      {
        assert( I->GetDataVect().size() == 2 );

        K   = new KleinWrapperObj;
        AND->PushBack(K);
        X   = new XeqObj;
        K->SetKleinChild(X);
        X->SetLink(I->GetDataVect()[1]);
        break;
      }

      case  PUMP_Token:
      {
        assert( I->GetDataVect().size() == 2 );
        assert( !G );

        G   = new GrammarObj;
        G->SetLabel(I->GetDataVect()[1]);
        G->SetIsPump(true);
        break;
      }

      case  STMNT_Token:
      {
        assert( I->GetDataVect().size() == 2 );

        OR  = new OrRuleObj;
        OR->SetLabel(I->GetDataVect()[1]);
        G->PushBack(OR);
        break;
      }

      case  AND_Token:
      {
        assert( I->GetDataVect().size() == 1 );

        AND   = new AndRuleObj;
        OR->PushBack(AND);
        break;
      }

      case  REF_CMND_Token:
      {
        assert( I->GetDataVect().size() );

        K   = new KleinWrapperObj;
        AND->PushBack(K);
        R   = new RefObj;
        K->SetKleinChild(R);
        R->SetLink(I->GetDataVect()[1]);
        break;
      }

      case  KLEIN_Token:
      {
        assert( I->GetDataVect().size() > 0 );

        if( I->GetDataVect().size() == 2 )
        {
          if( I->GetDataVect()[1] == "*" )
          {
            K->SetKlein(GrammarAbs::KlienStar);
          }
          else if( I->GetDataVect()[1] == "+" )
          {
            K->SetKlein(GrammarAbs::KlienPlus);
          }
          else if( I->GetDataVect()[1] == "?" )
          {
            K->SetKlein(GrammarAbs::KlienQuestion);
          }
          else
          {
            assert(false);
          }
        }
        else
        { // No Klein modifier --- we can remove the Klein node layer for speed
          GrammarAbs*   KleinSon  = K->GetKlienChild();
          AndRuleObj*   AndLayer  = dynamic_cast<AndRuleObj*>(K->GetFather());

          assert(AndLayer);

          K->ClearKleinChild();
          AndLayer->PopBack();
          delete    K;
          AndLayer->PushBack(KleinSon);
        }

        break;
      }

      case  OUT_LIST_START_Token:
      {
        assert( I->GetDataVect().size() == 2 );

        K   = new KleinWrapperObj;
        AND->PushBack(K);
        O   = new OutputListObj( OutputListObj::NewEntry );
        K->SetKleinChild(O);
        O->SetLabel(I->GetDataVect()[1]);

        break;
      }

      case  OUT_LIST_A_ITEM_Token:
      {
        assert( I->GetDataVect().size() == 1 );

        K   = new KleinWrapperObj;
        AND->PushBack(K);
        O   = new OutputListObj( OutputListObj::AppendItem );
        K->SetKleinChild(O);

        break;
      }

      case  OUT_LIST_A_CHAR_Token:
      {
        assert( I->GetDataVect().size() == 1 );

        K   = new KleinWrapperObj;
        AND->PushBack(K);
        O   = new OutputListObj( OutputListObj::AppendChar );
        K->SetKleinChild(O);

        break;
      }

      case  OUT_LIST_EL_Token:
      {
        assert( I->GetDataVect().size() == 3 );

        istringstream     iS(I->GetDataVect()[1]);
        int       II;
        iS >> II;
        assert(O);
        O->PushRuleNumBack( II, I->GetDataVect()[2] );

        break;
      }

      case  CHAR_START_Token:
      {
        assert( I->GetDataVect().size() == 1 );

        K   = new KleinWrapperObj;
        AND->PushBack(K);
        C   = new CharacterObj;
        K->SetKleinChild(C);

        break;
      }

      case  CHAR_Token:
      {
        assert( I->GetDataVect().size() == 2 );

        C2  = C1;
        C1  = I->GetDataVect()[1][0];

        break;
      }

      case  CHAR_SINGLE_Token:
      {
        assert( I->GetDataVect().size() == 1 );

        C->SetChar(C1);

        break;
      }

      case  CHAR_RANGE_Token:
      {
        assert( I->GetDataVect().size() == 1 );

        C->SetRangeChars( C2, C1 );

        break;
      }

      case  CHAR_ALL_Token:
      {
        assert( I->GetDataVect().size() == 1 );

        C->SetRangeChars( 0, 0xFF );

        break;
      }

      case  CHAREOF_Token:
      {
        assert( I->GetDataVect().size() == 1 );

        C2  = C1;
        C1  = CharacterObj::EOF_TOKEN;

        break;
      }

      case  CHAR_NOT_END_Token:
      {
        assert( I->GetDataVect().size() == 1 );

        C->Invert();

        break;
      }

      case  HEXCHAR_Token:
      {
        assert( I->GetDataVect().size() == 2 );

        istringstream     iS(I->GetDataVect()[1]);
        int       II;
        iS >> hex >> II;
        C2  = C1;
        C1  = II;

        break;
      }

      case  LITERAL_Token:
      {
        assert( I->GetDataVect().size() == 2 );

        K   = new KleinWrapperObj;
        AND->PushBack(K);
        L   = new LiteralObj;
        K->SetKleinChild(L);
        L->SetValue(I->GetDataVect()[1]);

        break;
      }

      case  LIT_START_Token:
      {
        assert( I->GetDataVect().size() == 1 );

        K   = new KleinWrapperObj;
        AND->PushBack(K);
        L   = new LiteralObj;
        K->SetKleinChild(L);
        Literal.clear();

        break;
      }

      case  LIT_EL_Token:
      {
        assert( I->GetDataVect().size() == 2 );

        Literal   += I->GetDataVect()[1][0];

        break;
      }

      case  LIT_HEX2_Token:
      {
        assert( I->GetDataVect().size() == 3 );

        Literal   += ((I->GetDataVect()[1][0] - '0') << 8) | (I->GetDataVect()[2][0] - '0');

        break;
      }

      case  LIT_END_Token:
      {
        assert( I->GetDataVect().size() == 1 );

        L->SetValue( Literal );

        break;
      }

      case  TAG_Token:
      {
        assert( I->GetDataVect().size() == 2 );

        K   = new KleinWrapperObj;
        AND->PushBack(K);
        T   = new TagObj;
        K->SetKleinChild(T);
        T->SetLabel(I->GetDataVect()[1]);

        break;
      }

      default:
      {
        cout << ">>UNKNOWN: '" << I->GetDataVect()[0] << "'" << endl;

        assert(false);
      }
    }
  }

  return  G;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// void   GrammarObj::GenerateCppParseTree
///////////////////////////////////////////////////////////////////////////////////////////////////

void    GrammarObj::GenerateCppParseTree( IBnfToCppCov& OutputData, bool KleinLayer, string& Includes )
{
  unsigned int                    K;
  IBnfToCppCov::SubRoutineType    ST;
  OrRuleObj*                      OR;
  ostringstream                   oS0;

  // Save label
  OutputData.GrammarLabel  = this->GetLabel();

  for( K=0 ; K < Group.size() ; ++K )
  {
    ostringstream   oS, oS2;

    OR   = dynamic_cast<OrRuleObj*>(Group[K]);

    ST.Label  = Group[K]->GetLabel();

    oS    << "GrammarAbs*    Parse_" << ST.Label << "(void)" << endl;
    oS    << "{" << endl;
    oS    << "  ///////////////////////////////////////////////////////////" << endl;
    oS    << "  // OR Group '" << ST.Label << "'" << endl;
    oS    << "  ///////////////////////////////////////////////////////////" << endl;
    oS    << endl;
    oS    << "  OR  = new OrRuleObj;" << endl;
    oS    << "  OR->SetLabel(\"" << ST.Label << "\");" << endl;

    // Save this record
    OutputData.Subroutines.push_back(ST);

    // Accumulate the output
    OutputData.Subroutines.back().Contents  += oS.str();

    // Process the sub-rules
    OR->GenerateCppParseTree( OutputData, KleinLayer, Includes );

    oS2   << endl;
    oS2   << "  return    OR;" << endl;
    oS2   << "}" << endl;
    oS2   << endl;
    oS2   << endl;

    // Accumulate the output
    OutputData.Subroutines.back().Contents  += oS2.str();
  }

  ST.Label  = this->GetLabel();

  oS0   << "// Variables used in subroutines" << endl;
  oS0   << "GrammarObj*       GRAM;" << endl;
  oS0   << "CharacterObj*     CHAR;" << endl;
  oS0   << "KleinWrapperObj*  KLEIN;" << endl;
  oS0   << "RefObj*           REF;" << endl;
  oS0   << "LiteralObj*       LIT;" << endl;
  oS0   << "OutputListObj*    OUTL;" << endl;
  oS0   << "AndRuleObj*       AND;" << endl;
  oS0   << "OrRuleObj*        OR;" << endl;
  oS0   << endl;

  oS0   << "// Function Prototypes" << endl;

  // Call all of the subroutines
  for( K = 0 ; K < Group.size() ; ++K )
  {
    oS0 << "GrammarAbs*    Parse_" << Group[K]->GetLabel() << "(void);" << endl;
  }

  // Get the includes
  oS0 << endl;
  oS0 << "// XEQ Includes" << endl;
  oS0 << Includes << endl;

  oS0   << endl;
  oS0   << "GrammarAbs*   Parse_" << OutputData.GrammarLabel << "(void)" << endl;
  oS0   << "{" << endl;
  oS0   << "  ///////////////////////////////////////////////////////////" << endl;
  oS0   << "  // GRAM Group" << endl;
  oS0   << "  ///////////////////////////////////////////////////////////" << endl;
  oS0   << endl;
  oS0   << "  GRAM  = new GrammarObj;" << endl;
  oS0   << "  GRAM->SetLabel(\"" << OutputData.GrammarLabel << "\");" << endl;
  oS0   << endl;
  oS0   << "  ///////////////////////////////////////////////////////////" << endl;
  oS0   << "  // Rule Entry Points" << endl;
  oS0   << "  ///////////////////////////////////////////////////////////" << endl;
  oS0   << endl;

  // Call all of the subroutines
  for( K = 0 ; K < Group.size() ; ++K )
  {
    oS0 << "  GRAM->PushBack(Parse_" << Group[K]->GetLabel() << "());" << endl;
  }

  oS0   << endl;
  oS0   << "  return  GRAM;" << endl;
  oS0   << "}" << endl;
  oS0   << endl;
  oS0   << endl;

  // Save this record
  OutputData.Subroutines.push_front(ST);

  // Accumulate the output
  OutputData.Subroutines.front().Contents  += oS0.str();

  oS0 << endl;
}


///////////////////////////////////////////////////////////////////////////////////////////////////
// OrRuleObj*    GrammarObj::LookUpRuleByLabel
///////////////////////////////////////////////////////////////////////////////////////////////////

OrRuleObj*    GrammarObj::LookUpRuleByLabel( const string LabelToLookup )
{
  map<const string,MapRecType>::const_iterator    I   = Rules.find(LabelToLookup);
  GrammarAbs*                                     GA  = I->second.Address;
  OrRuleObj*                                      OR  = dynamic_cast<OrRuleObj*>(GA);
  return      OR;
}
}   // End namespace 'wej_parser'


