///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  ComplexHermitionVector.h                                                          //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#pragma once

#include  <cassert>
#include  <cstdint>
#include  <algorithm>
#include  "Exception.h"
#include  "AlignedMemAlloc.h"

using namespace std;

///////////////////////////////////////////////////////////////////////////////////////////////////
//  ComplexHermitionVector
///////////////////////////////////////////////////////////////////////////////////////////////////
//
//  A ComplexHermitionVector is a conjugate-symmetric vector that is uniquely represented by its
//       first '(N>>1) + 1' components.  The inverse size conversion is given by '((N>>1)+1)>>1'.
//
///////////////////////////////////////////////////////////////////////////////////////////////////

template<typename Type>
class  ComplexHermitionVector
{
private:

  AVX_Vector<Type>  Vector;
  uint32_t          FullLength;

public:

  ComplexHermitionVector( const uint32_t _FullLength, Type _Fill=0 ) :
  Vector( this->ConvFullToFoldedLength(_FullLength), _Fill ),
  FullLength(_FullLength)
  {
  }

  // Assuming '_FullConjSymArray' is conjugate-symmetric
  const ComplexHermitionVector& ConvFullToFolded( const AVX_Vector<Type>& _FullConjSymArray )
  {
    const uint32_t  VectorEnd = this->ConvFullToFoldedLength(_FullConjSymArray.size());

    if( VectorEnd != this->Vector.size() )
    {
      this->Vector.resize(VectorEnd);
    }

    std::copy
    (
      &_FullConjSymArray[0],
      &_FullConjSymArray[VectorEnd],
      &this->Vector[0]
    );

    return  *this;
  }

  //////////
  // Sizes
  //////////

  static uint32_t  ConvFullToFoldedLength( const unsigned int _FullLength )
  {
    return  (_FullLength >> 1) + 1;
  }

  uint32_t  GetFullLength(void) const
  {
    return  this->FullLength;
  }

  uint32_t  GetFoldedLength(void) const
  {
    return  this->Vector.size();
  }

  uint32_t  ConvPosFullToFolded( const uint32_t _Pos ) const
  {
    if( _Pos >= this->GetFullLength() )
    {
      Exception  E(__FILE__,__LINE__);
      E.Display() = "ComplexHermitionVector: 'ConvPosFullToFolded' accessed out of range.";
      throw  E;
    }

    if( _Pos < this->GetFoldedLength() )
    {
      return  _Pos;
    }
    else
    {
      return  (this->FullLength & ~1) - _Pos;
    }
  }

  //////////////////////
  // Element Accessors
  //////////////////////

  const Type  operator[]( const uint32_t _Pos ) const
  {
    if( _Pos >= this->GetFullLength() )
    {
      Exception  E(__FILE__,__LINE__);
      E.Display() = "ComplexHermitionVector: 'operator[]' accessed out of range.";
      throw  E;
    }

    const uint32_t  FoldedPos = this->ConvPosFullToFolded(_Pos);
    const bool      HasFolded = _Pos >= this->GetFoldedLength();

    if( !HasFolded )
    {
      return  this->Vector[_Pos];
    }
    else
    {
      return  conj(this->Vector[FoldedPos]);
    }
  }

  // This function can create non conjugate-symmetric vectors, which are assumed
  // in processing,
  void  Set( const uint32_t _Pos, const Type& _Data )
  {
    if( _Pos >= this->GetFullLength() )
    {
      Exception  E(__FILE__,__LINE__);
      E.Display() = "ComplexHermitionVector: 'Set' accessed out of range.";
      throw  E;
    }

    const uint32_t  FoldedPos = this->ConvPosFullToFolded(_Pos);
    const bool      HasFolded = _Pos >= this->GetFoldedLength();

    if( !HasFolded )
    {
      this->Vector[_Pos] = _Data;
    }
    else
    {
      this->Vector[FoldedPos] = conj(_Data);
    }
  }

  //////////////
  // Operators
  //////////////

  ComplexHermitionVector  operator+( const ComplexHermitionVector& _Other ) const
  {
    if( this->Vector.size() != _Other.Vector.size() )
    {
      Exception  E(__FILE__,__LINE__);
      E.Display() = "ComplexHermitionVector: 'operator+' vectors of differing sizes";
      throw  E;
    }

    ComplexHermitionVector  CHV( this->Vector.size() );

    for( auto K=0U ; K < this->Vector.size() ; ++K )
    {
      CHV.Vector[K] = this->Vector[K] + _Other.Vector[K];
    }

    return  CHV;
  }

  ComplexHermitionVector  operator-( const ComplexHermitionVector& _Other ) const
  {
    if( this->Vector.size() != _Other.Vector.size() )
    {
      Exception  E(__FILE__,__LINE__);
      E.Display() = "ComplexHermitionVector: 'operator-' vectors of differing sizes";
      throw  E;
    }

    ComplexHermitionVector  CHV( this->Vector.size() );

    for( auto K=0U ; K < this->Vector.size() ; ++K )
    {
      CHV.Vector[K] = this->Vector[K] - _Other.Vector[K];
    }

    return  CHV;
  }

  ComplexHermitionVector  operator*( const ComplexHermitionVector& _Other ) const
  {
    if( this->Vector.size() != _Other.Vector.size() )
    {
      Exception  E(__FILE__,__LINE__);
      E.Display() = "ComplexHermitionVector: 'operator*' vectors of differing sizes";
      throw  E;
    }

    ComplexHermitionVector  CHV( this->Vector.size() );

    for( auto K=0U ; K < this->Vector.size() ; ++K )
    {
      CHV.Vector[K] = this->Vector[K] * _Other.Vector[K];
    }

    return  CHV;
  }

  ComplexHermitionVector  operator/( const ComplexHermitionVector& _Other ) const
  {
    if( this->Vector.size() != _Other.Vector.size() )
    {
      Exception  E(__FILE__,__LINE__);
      E.Display() = "ComplexHermitionVector: 'operator/' vectors of differing sizes";
      throw  E;
    }

    ComplexHermitionVector  CHV( this->Vector.size() );

    for( auto K=0U ; K < this->Vector.size() ; ++K )
    {
      CHV.Vector[K] = this->Vector[K] / _Other.Vector[K];
    }

    return  CHV;
  }

  const ComplexHermitionVector& operator*=( const ComplexHermitionVector& _Other )
  {
    if( this->Vector.size() != _Other.Vector.size() )
    {
      Exception  E(__FILE__,__LINE__);
      E.Display() = "ComplexHermitionVector: 'operator*=' vectors of differing sizes";
      throw  E;
    }

    for( auto K=0U ; K < this->Vector.size() ; ++K )
    {
      this->Vector[K] = this->Vector[K] * _Other.Vector[K];
    }

    return  *this;
  }

  ////////////////
  // Conversions
  ////////////////

  AVX_Vector<Type>&  AccessToFoldedVector(void)
  {
    return  this->Vector;
  }

  AVX_Vector<Type>  ConvetToStandardHermitionVector(void) const
  {
    AVX_Vector<Type>  Out(this->GetFullLength());

    for( auto K=0U ; K < this->GetFullLength() ; ++K )
    {
      Out[K] = this->operator[](K);
    }

    return  Out;
  }
};


