///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//  FilePath:  TemplateRenderer.cpp                                                              //
//  Author:    William Earl Jones                                                                //
//  Email:     wejc@wejc.com                                                                     //
//  Tag:       DEVELOP                                                                           //
//  TimeDate:  20231110_143153                                                                   //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////////////
//                                                                                               //
//                           Copyright (c) 2023 William Earl Jones                               //
//                                                                                               //
//  Redistribution and use in source and binary forms, with or without modification, are         //
//  permitted provided that the following conditions are met:                                    //
//                                                                                               //
//    1. Redistributions of source code must retain the above copyright notice, this list of     //
//       conditions and the following disclaimer.                                                //
//                                                                                               //
//    2. Redistributions in binary form must reproduce the above copyright notice, this list of  //
//       conditions and the following disclaimer in the documentation and/or other materials     //
//       provided with the distribution.                                                         //
//                                                                                               //
//    3. Neither the name of the copyright holder nor the names of its contributors may be used  //
//       to endorse or promote products derived from this software without specific prior        //
//       written permission.                                                                     //
//                                                                                               //
//  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS  //
//  OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF              //
//  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE   //
//  COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,    //
//  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF           //
//  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)       //
//  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR     //
//  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, //
//  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.                                           //
//                                                                                               //
///////////////////////////////////////////////////////////////////////////////////////////<wejc>//

#include    <map>
#include    "PARSER_Template.h"
#include    "DataRef.h"
#include    "TemplateRenderer.h"


wej_parser::GrammarAbs*   TemplateRenderer::GrammarConverter = nullptr;


TemplateRenderer::TemplateRenderer(void)
{
  if( GrammarConverter == nullptr )
  { // Recover the parse tree generator routine
    GrammarConverter =  wej_parser::wej_template::Parse_Path();

    if(!GrammarConverter)
    {
      throw  Exception("Cannot run 'wej_parser::wej_template::Parse_Path'.",__FILE__,__LINE__);
    }
  }
}


bool    TemplateRenderer::Parse( istream& In )
{
  bool  R = false;

  if(!GrammarConverter)
  {
    throw  Exception("Cannot run 'wej_parser::wej_template::Parse_Path'.",__FILE__,__LINE__);
  }

  Outputs.Clear();

  wej_parser::ParseFrame   Frame( In, Outputs );

  R  = GrammarConverter->Parse(Frame);

  Frame.AcceptAll();

  if(!R)
  {
    throw  Exception("Cannot parse Template file.",__FILE__,__LINE__);
  }

  // Display Outputs [WEJTAG_OKAY,DEBUG]
  if(false)
  {
    cout << endl << Outputs.Display() << endl;
  }

  return    R;
}


string  TemplateRenderer::ProcessOutput( const DataPath& XmlProcRoot )
{
  ostringstream   Out;

  IterStack.clear();

  IterStack.push_back( StackRec( XmlProcRoot, Outputs.Begin() ) );

  while( ProcessNextStep( Out, XmlProcRoot ) );

  return  Out.str();
}


const map<string,int>   TemplateRenderer::TokenToRuleMap
{
  { "ORIG",     TemplateRenderer::ORIGINAL_Token },
  { "VAL",      TemplateRenderer::VAL_Token      },
  { "BARVAL",   TemplateRenderer::BAR_VAL_Token  },
  { "ITER",     TemplateRenderer::ITER_Token     },
  { "END_ITER", TemplateRenderer::END_ITER_Token },
  { "BTW",      TemplateRenderer::BTW_Token      },
  { "STR",      TemplateRenderer::STR_Token      }
};


bool  TemplateRenderer::ProcessNextStep( ostringstream& Out, const DataPath& XmlProcRoot )
{
  // Quick rule lookup
  auto  I_Rule = TokenToRuleMap.find( IterStack.back().GetOut()->GetDataVect()[0] );

  if( I_Rule == TokenToRuleMap.end() )
  {
    throw  Exception(string("Cannot find tag '" + IterStack.back().GetOut()->GetDataVect()[0] + "'"),__FILE__,__LINE__);
  }

  switch(I_Rule->second)
  {
    case  ORIGINAL_Token:
    {
      Out << IterStack.back().GetOut()->GetDataVect()[1];
      IterStack.back().BumpOutput();
    }
    break;

    case  VAL_Token:
    {
      DataPath  DP;

      // If there are children then we look them up and add their output else ignore
      if( !IterStack.back().GetDNR().IsRootLeaf() )
      {
        // If the stack depth is greater than 1 then we use the 'Curr' else we don't use the child
        if( IterStack.size() == 1 )
        {
          // Form path to XML VAL Token - Post-pend the new node
          DP = IterStack.back().GetDNR().GetDataPath() +
               IterStack.back().GetOut()->GetDataVect()[1];
        }
        else
        {
          // Form path to XML VAL Token - Post-pend the new node
          DP = IterStack.back().GetDNR().Curr().GetDataPath() +
               IterStack.back().GetOut()->GetDataVect()[1];
        }

        Out << LookUpWithIndirection( DP, XmlProcRoot );
      }

      IterStack.back().BumpOutput();
    }
    break;

    // A VAL command that DOES NOT use the iterator
    case  BAR_VAL_Token:
    {
      DataPath    DP = IterStack.front().GetDNR().GetDataPath()  +
                       IterStack.back().GetOut()->GetDataVect()[1];

      Out << LookUpWithIndirection( DP, XmlProcRoot );
      IterStack.back().BumpOutput();
    }
    break;

    case  STR_Token:
    {
      CurrentString = IterStack.back().GetOut()->GetDataVect()[1];
      IterStack.back().BumpOutput();
    }
    break;

    case  BTW_Token:
    {
      // See if we are in an ITER Loop
      if( IterStack.size() > 1 )
      { // In ITER Loop --- See if this is the last loop
        auto  LocDNR = IterStack.back().GetDNR();

        LocDNR.Next();

        if( LocDNR.MoreChildren() )
        { // Not the last loop.
          Out << CurrentString;
        }
      }
      else
      { // Not in ITER Loop
        Out << CurrentString;
      }

      IterStack.back().BumpOutput();
    }
    break;

    case  ITER_Token:
    {
      if( !IterStack.back().GetDNR().MoreChildren() )
      {
        throw  Exception(string("Attempt to access sublist of node without children."),__FILE__,__LINE__);
      }

      StackRec&  SR0 = IterStack.back();

      // Path to node whose children we iterate over
      DataPath  DP;

      // If the stack depth is greater than 1 then we use the 'Curr' else we don't use the child
      if( IterStack.size() == 1 )
      {
        // Form path to XML VAL Token - Post-pend the new node
        DP = SR0.GetDNR().GetDataPath() + SR0.GetOut()->GetDataVect()[1];
      }
      else
      {
        // Form path to XML VAL Token - Post-pend the new node
        DP = SR0.GetDNR().Curr().GetDataPath() + SR0.GetOut()->GetDataVect()[1];
      }

      IterStack.push_back( StackRec( DP, SR0.GetOut() ) );


//  cout << "<ITER_Token 1> DP:" << DP.DisplayAll(2) << endl;  cout.flush();


      IterStack.back().BumpOutput();
    }
    break;

    case  END_ITER_Token:
    {
      // Locate the level below
      auto         Rev   = IterStack.rbegin();
      const auto&  Lvl_1 = *(++Rev);
      StackRec&    SR0   = IterStack.back();

      // If there are children then we are 'At End'
      if( SR0.GetDNR().IsRootLeaf() )
      { // At end
        auto  OutI = SR0.GetOut();
        IterStack.pop_back();
        IterStack.back().SetOut( ++OutI );
        break;
      }

      // Point to the next child
      SR0.GetDNR().Next();

      if( SR0.GetDNR().MoreChildren() )
      { // Not at end --- Point to the next label in the list
        SR0.SetOut( Lvl_1.GetOut() );
        SR0.SetOut( ++IterStack.back().GetOut() );
      }
      else
      { // 'At End'
        auto  OutI = SR0.GetOut();
        IterStack.pop_back();
        IterStack.back().SetOut( ++OutI );
      }
    }
    break;

    default:
    {
      throw  Exception(string("Cannot find tag '" +
               IterStack.back().GetOut()->GetDataVect()[0] + "'"),__FILE__,__LINE__);
    }
    break;
  }

  return  IterStack.back().GetOut() != Outputs.GetDataList().end();
}


// If NO Indirection ignore _RootPath.
// If    Indirection then lookup with respect to _RootPath
std::string  TemplateRenderer::LookUpWithIndirection
                ( const DataPath& XML_Path, const DataPath& RootPath, bool NoIndirection )
{


/*
cout << "<TemplateRenderer::LookUpWithIndirection> " << XML_Path.GetPathString() << endl
     << "  " << XML_Path.GetPathString() << endl
     << "  " << RootPath.GetPathString() << endl
     << "  " << "NoIndirection=" << boolalpha << NoIndirection << endl;
cout.flush();
*/


  DataRef<string>  DR( XML_Path );

//  cout << "<TemplateRenderer::DR0> " << DR.GetDataAsString() << endl;  cout.flush();

  if(NoIndirection)
  {

//  cout << "<TemplateRenderer DIRECT> " << DR.GetDataAsString() << endl;  cout.flush();

    return  DR.GetDataAsString();
  }
  else
  {


//  cout << "<TemplateRenderer INDIRECT> " << boolalpha << DR.IsPossibleIndirection() << endl;  cout.flush();


    while( DR.IsPossibleIndirection() )
    {


//  cout << "<TemplateRenderer INDIRECT-pre> " << RootPath << endl;  cout.flush();


      DR.PerformIndirection( RootPath );


// cout << "<TemplateRenderer INDIRECT-post> "  << endl;  cout.flush();


    }
  }

// cout << "<TemplateRenderer::DR9> " << DR.GetDataAsString() << endl;  cout.flush();

  return  DR.GetDataAsString();
}


